﻿Imports System.IO
Imports System.Diagnostics

Public Class HpglViewer

    Structure Line
        Dim pen As Integer
        Dim points As List(Of Point)
    End Structure

    Enum Action
        Unknown
        Initialize
        Scale
        SelectPen
        PenUp
        PenDown
        PlotAbsolute
        PlotRelative
    End Enum

    Structure Command
        Dim command As Action
        Dim parameters As List(Of Integer)
    End Structure

    Dim logWindow As DataLog = Nothing
    Dim filePath As String = ""
    Dim WithEvents serialPort As Ports.SerialPort = Nothing
    Dim receivingData As Boolean = False
    Dim currentFile(1024 * 1024) As Byte
    Dim currentFileIndex As Integer
    Dim incomingData(1024) As Byte
    Dim incomingDataCount As Integer

    Dim themeWhite As Color() = {Color.White, Color.Black, Color.Red, Color.Green, Color.Blue, Color.Cyan, Color.Magenta, Color.Yellow}
    Dim themeDark As Color() = {Color.Black, Color.Gray, Color.Yellow, Color.Cyan, Color.Magenta, Color.Blue, Color.Green, Color.Red}
    Dim keepAspect As Boolean = True

    Dim canvasWidth As Integer
    Dim canvasHeight As Integer

    Dim plotDimensions As Integer()
    Dim plotScale As Integer()
    Dim plotElements As New List(Of Line)
    Dim plotTheme As Color() = themeWhite
    Dim plotBitmap As Bitmap
    Dim plotRotate As Integer = 0

    Private Sub HpglViewer_Load(sender As Object, e As EventArgs) Handles MyBase.Load
        PlotInitialize()
        LabelStatus.Text = ""
        OpenFileDialog1.InitialDirectory = "C:\"

        For Each item As String In My.Computer.Ports.SerialPortNames
            Dim port As ToolStripMenuItem = New ToolStripMenuItem(item)
            port.Name = item
            AddHandler port.Click, AddressOf SerialPortSelected
            PortToolStripMenuItem.DropDownItems.Add(port)
        Next

        logWindow = New DataLog()
    End Sub

    Private Sub SerialPortSelected(ByVal sender As System.Object, ByVal e As System.EventArgs)

        ' Get port name
        Dim portName As String = DirectCast(sender, ToolStripMenuItem).Name

        ' Clear all checkmarks
        For Each item As ToolStripMenuItem In PortToolStripMenuItem.DropDownItems
            item.Checked = False
        Next

        ' See what to do about current serial port
        If serialPort IsNot Nothing Then
            If serialPort.PortName = portName Then
                serialPort.Close()
                LabelStatus.Text = "Closed serial port " & portName
                Exit Sub
            Else
                serialPort.Close()
            End If
        End If

        ' Open new serial port
        Try
            serialPort = My.Computer.Ports.OpenSerialPort(portName)
            serialPort.BaudRate = 460800
            DirectCast(sender, ToolStripMenuItem).Checked = True
            LabelStatus.Text = "Opened serial port " & portName
        Catch
            serialPort = Nothing
            LabelStatus.Text = "Could not open serial port " & portName
        End Try

    End Sub

    Private Sub serialPort_DataReceived(ByVal sender As Object, ByVal e As System.IO.Ports.SerialDataReceivedEventArgs) Handles serialPort.DataReceived
        Try
            If serialPort.IsOpen Then
                incomingDataCount = serialPort.Read(incomingData, 0, incomingData.Length())
                Invoke(New EventHandler(AddressOf ProcessReceivedData))
            End If
        Catch ex As Exception
        End Try
    End Sub

    Private Sub ProcessReceivedData()
        If receivingData = False Then
            currentFileIndex = 0
            receivingData = True
        End If
        Array.Copy(incomingData, 0, currentFile, currentFileIndex, incomingDataCount)
        currentFileIndex = currentFileIndex + incomingDataCount
        LabelStatus.Text = "Received " & currentFileIndex & " bytes"
    End Sub

    Private Sub Timer1_Tick(sender As Object, e As EventArgs) Handles Timer1.Tick
        Static Dim previousFileIndex As Integer = 0
        If receivingData = True Then
            If previousFileIndex = currentFileIndex Then
                LabelStatus.Text = "No more incoming data, hardcopy finished"
                PlotInitialize()
                ParseData(System.Text.Encoding.ASCII.GetString(currentFile))
                PaintPlot()
                PanelCanvas.Invalidate()
                receivingData = False
            Else
                previousFileIndex = currentFileIndex
            End If
        End If
    End Sub

    Private Sub PlotInitialize()
        plotDimensions = {Integer.MaxValue, 0, Integer.MaxValue, 0}
        plotScale = {0, 0, 0, 0}
        plotElements.Clear()
    End Sub

    Private Sub PlotLine(pen As Integer, startPoint As Point, endPoint As Point)
        Dim l As Line
        l.pen = pen
        l.points = New List(Of Point)
        l.points.Add(startPoint)
        l.points.Add(endPoint)
        plotElements.Add(l)
    End Sub

    Private Function ParseHpgl(s As String) As Command
        Dim r As Command = New Command

        r.command = Action.Unknown
        r.parameters = New List(Of Integer)

        If s.Length < 2 Then
            GoTo CANCEL
        End If

        Dim directive As String = s.Substring(0, 2)

        Select Case directive
            Case "IN" : r.command = Action.Initialize
            Case "SC" : r.command = Action.Scale
            Case "SP" : r.command = Action.SelectPen
            Case "PU" : r.command = Action.PenUp
            Case "PD" : r.command = Action.PenDown
            Case "PA" : r.command = Action.PlotAbsolute
            Case "PR" : r.command = Action.PlotRelative
            Case Else
                Debug.Print("Unknown directive: " + directive)
                r.command = Action.Unknown
        End Select

        s = s.Remove(0, 2)

        Dim parameters = s.Split(New Char() {","c})
        Dim parameter As String = ""

        For Each parameter In parameters
            Dim value As Integer = 0
            If Integer.TryParse(parameter, value) = True Then
                r.parameters.Add(value)
            End If
        Next

CANCEL:
        Return r
    End Function

    Private Sub OpenFile(fileName As String)

        Try
            currentFile = File.ReadAllBytes(fileName)
            LabelStatus.Text = "Opened file " & fileName
        Catch
            LabelStatus.Text = "Could not open file " & fileName
            Exit Sub
        End Try

        PlotInitialize()
        ParseData(System.Text.Encoding.ASCII.GetString(currentFile))
        PaintPlot()
        PanelCanvas.Invalidate()

    End Sub

    Private Sub ParseData(s As String)

        ' Plotter state variables
        Dim currentPos As Point = New Point(0, 0)
        Dim penDown As Boolean = False
        Dim penColor As Integer = 1
        Dim plotRelative As Boolean = False
        Dim plotScaleSet As Boolean = False

        ' Loop through file
        Dim commands As String() = s.Split(New Char() {";"c})
        Dim token As String

        For Each token In commands
            Dim instruction As Command = ParseHpgl(token)

            Select Case instruction.command
                Case Action.Initialize
                    If instruction.parameters.Count = 0 Then
                        PlotInitialize()
                    End If

                Case Action.SelectPen
                    If instruction.parameters.Count = 1 Then
                        penColor = instruction.parameters(0)
                    End If

                Case Action.Scale
                    If instruction.parameters.Count = 4 Then
                        plotScale = instruction.parameters.ToArray()
                        plotScaleSet = True
                    End If

                Case Action.PenDown
                    penDown = True
                    If instruction.parameters.Count = 2 Then
                        Dim newPos As Point = New Point(instruction.parameters(0), instruction.parameters(1))
                        PlotLine(penColor, currentPos, newPos)
                        currentPos = newPos
                    End If

                Case Action.PenUp
                    penDown = False
                    If instruction.parameters.Count = 2 Then
                        Dim newPos As Point = New Point(instruction.parameters(0), instruction.parameters(1))
                        currentPos = newPos
                    End If

                Case Action.PlotAbsolute
                    plotRelative = False
                    If instruction.parameters.Count = 2 Then
                        Dim newPos As Point = New Point(instruction.parameters(0), instruction.parameters(1))
                        If penDown = True Then PlotLine(penColor, currentPos, newPos)
                        currentPos = newPos
                    End If

                Case Action.PlotRelative
                    plotRelative = True
                    If instruction.parameters.Count = 2 Then
                        Dim newPos As Point = New Point(currentPos.X + instruction.parameters(0), currentPos.Y + instruction.parameters(1))
                        If penDown = True Then PlotLine(penColor, currentPos, newPos)
                        currentPos = newPos
                    End If

            End Select

            ' Keep track of coordinate range
            If currentPos.X < plotDimensions(0) Then plotDimensions(0) = currentPos.X
            If currentPos.X > plotDimensions(1) Then plotDimensions(1) = currentPos.X
            If currentPos.Y < plotDimensions(2) Then plotDimensions(2) = currentPos.Y
            If currentPos.Y > plotDimensions(3) Then plotDimensions(3) = currentPos.Y
        Next

        ' If no scale was set in the file, determine size from coordinates
        If plotScaleSet = False Then
            plotScale(1) = plotDimensions(1) - plotDimensions(0)
            plotScale(3) = plotDimensions(3) - plotDimensions(2)
        End If

    End Sub

    Private Sub PaintPlot()

        canvasWidth = PanelCanvas.Width
        canvasHeight = PanelCanvas.Height - PanelStatus.Height

        If canvasWidth <= 0 Or canvasHeight <= 0 Then Exit Sub
        If plotScale Is Nothing Then Exit Sub
        If plotScale.Count < 4 Then Exit Sub

        Dim plotWidth As Integer = plotScale(1) - plotScale(0)
        Dim plotHeight As Integer = plotScale(3) - plotScale(2)

        If plotWidth <= 0 Or plotHeight <= 0 Then Exit Sub

        Dim canvasAspect As Double = canvasWidth / canvasHeight
        Dim plotAspect As Double = plotWidth / plotHeight

        Debug.Print(String.Format("plotWidth: {0}, plotHeight: {1}, plotAspect: {2}, canvasAspect: {3}", plotWidth, plotHeight, plotAspect, canvasAspect))

        If keepAspect = True Then
            If plotAspect > canvasAspect Then
                canvasHeight = canvasWidth / plotAspect
            Else
                canvasWidth = canvasHeight * plotAspect
            End If
        End If

        Dim widthscale As Double = canvasWidth / plotWidth
        Dim heightScale As Double = canvasHeight / plotHeight

        Debug.Print(String.Format("widthScale: {0}, heightScale: {1}", widthScale, heightScale))

        ' Initialize suitable bitmap for drawing
        plotBitmap = New Bitmap(canvasWidth, canvasHeight, Imaging.PixelFormat.Format24bppRgb)
        Dim graphics As Graphics = graphics.FromImage(plotBitmap)
        graphics.Clear(plotTheme(0))
        graphics.SmoothingMode = Drawing2D.SmoothingMode.AntiAlias

        ' Draw all elements
        Dim pen As Pen = New Pen(Color.Black)
        pen.StartCap = Drawing2D.LineCap.Round
        pen.EndCap = Drawing2D.LineCap.Round
        Dim line As Line
        For Each line In plotElements
            Dim pointFrom As Point = New Point(line.points(0).X * widthScale, line.points(0).Y * heightScale)
            Dim pointTo As Point = New Point(line.points(1).X * widthscale, line.points(1).Y * heightScale)
            pen.Color = plotTheme(line.pen)
            graphics.DrawLine(pen, pointFrom, pointTo)
        Next

        ' Graphics coordinates are vertically mirrored from plot coordinates
        plotBitmap.RotateFlip(RotateFlipType.RotateNoneFlipY)

    End Sub

    Private Sub PanelCanvas_Paint(sender As Object, e As PaintEventArgs) Handles PanelCanvas.Paint
        If plotBitmap Is Nothing Then
            Exit Sub
        Else
            Dim graphics As Graphics = PanelCanvas.CreateGraphics()
            Dim offset As Point = New Point((PanelCanvas.Width - plotBitmap.Width) / 2, (PanelCanvas.Height - plotBitmap.Height - PanelStatus.Height) / 2)
            graphics.DrawImage(plotBitmap, offset)
        End If
    End Sub

    Private Sub HPGL_Resize(sender As Object, e As EventArgs) Handles MyBase.Resize
        PaintPlot()
        PanelCanvas.Invalidate()
    End Sub

    Private Sub OpenToolStripMenuItem_Click(sender As Object, e As EventArgs) Handles OpenToolStripMenuItem.Click
        OpenFileDialog1.Title = "Open File"
        OpenFileDialog1.Filter = "All files (*.*)|*.*|All files (*.*)|*.*"
        OpenFileDialog1.FilterIndex = 2

        If OpenFileDialog1.ShowDialog() = DialogResult.OK Then
            filePath = OpenFileDialog1.FileName
            PlotInitialize()
            OpenFile(filePath)
            PaintPlot()
            PanelCanvas.Invalidate()
        End If
    End Sub

    Private Sub WhiteToolStripMenuItem_Click(sender As Object, e As EventArgs) Handles WhiteToolStripMenuItem.Click
        If WhiteToolStripMenuItem.Checked = False Then
            WhiteToolStripMenuItem.Checked = True
            BlackToolStripMenuItem.Checked = False
            plotTheme = themeWhite
            PaintPlot()
            PanelCanvas.Invalidate()
        End If
    End Sub

    Private Sub BlackToolStripMenuItem_Click(sender As Object, e As EventArgs) Handles BlackToolStripMenuItem.Click
        If BlackToolStripMenuItem.Checked = False Then
            BlackToolStripMenuItem.Checked = True
            WhiteToolStripMenuItem.Checked = False
            plotTheme = themeDark
            PaintPlot()
            PanelCanvas.Invalidate()
        End If
    End Sub

    Private Sub X480ToolStripMenuItem_Click(sender As Object, e As EventArgs) Handles X480ToolStripMenuItem.Click
        Me.Width = 640
        Me.Height = 480
        PaintPlot()
        PanelCanvas.Invalidate()
    End Sub

    Private Sub X600ToolStripMenuItem_Click(sender As Object, e As EventArgs) Handles X600ToolStripMenuItem.Click
        Me.Width = 800
        Me.Height = 600
        PaintPlot()
        PanelCanvas.Invalidate()
    End Sub

    Private Sub X768ToolStripMenuItem_Click(sender As Object, e As EventArgs) Handles X768ToolStripMenuItem.Click
        Me.Width = 1024
        Me.Height = 768
        PaintPlot()
        PanelCanvas.Invalidate()
    End Sub

    Private Sub RotateCwToolStripMenuItem_Click(sender As Object, e As EventArgs) Handles RotateCwToolStripMenuItem.Click

    End Sub

    Private Sub RotateCcwToolStripMenuItem_Click(sender As Object, e As EventArgs) Handles RotateCcwToolStripMenuItem.Click

    End Sub

    Private Sub MirrorToolStripMenuItem_Click(sender As Object, e As EventArgs) Handles MirrorToolStripMenuItem.Click

    End Sub

    Private Sub FlipToolStripMenuItem_Click(sender As Object, e As EventArgs) Handles FlipToolStripMenuItem.Click

    End Sub

    Private Sub CommunicationToolStripMenuItem_Click(sender As Object, e As EventArgs) Handles PortToolStripMenuItem.Click

    End Sub

    Private Sub RawDataToolStripMenuItem_Click(sender As Object, e As EventArgs) Handles RawDataToolStripMenuItem.Click
        logWindow.Show()
    End Sub

End Class
